import 'dart:io';
import 'dart:async';
import 'package:adeptutils/adeptutils.dart' as AdeptUtils;
import 'package:chopper/chopper.dart';
import 'package:adeptarch_base/src/api/adept_chopper_client.dart';
import 'package:http/http.dart' as http;

final restAPI = RestAPI();

class RestAPI {
  http.Client? httpClient;
  AdeptChopperClient? chopperClient;
  Converter? converter;
  final Map<Type, ChopperService> _services = {};

  Duration? timeout;
  String? apiProtocol;
  String? apiHost;
  String? apiPort;
  late String apiUrl;

  RestAPI() {
    apiProtocol = "http://";
    apiHost = "www.myapi.com";
    apiPort = "8080";
    apiUrl = apiProtocol! + apiHost! + apiPort!;
  }

  void init({String? apiUrl, Duration? timeout, http.Client? httpClient}) {
    if (AdeptUtils.TextUtils.isNotEmpty(apiUrl)) updateApiUrl(apiUrl: apiUrl);
    if (timeout != null) this.timeout = timeout;
    if (httpClient != null) this.httpClient = httpClient;
    try {
      chopperClient?.dispose();
    } catch (e) {
      print(e);
    }
    converter ??= const JsonConverter();
    chopperClient = AdeptChopperClient(
      baseUrl: this.apiUrl,
      client: this.httpClient,
      services: _services.values,
      converter: converter,
      timeout: this.timeout,
    );
    _initServices();
  }

  void _initServices() {
    _services.forEach((type, service) {
      service.client = chopperClient!;
    });
  }

  void dispose() {
    try {
      // _services.forEach((t, s) => s.dispose()); // ChopperService no longer support .dispose()
      _services.clear();
      chopperClient?.dispose();
    } catch (e) {
      print(e);
    }
  }

  void addService(ChopperService service) {
    service.client = chopperClient!;
    _services[service.definitionType] = service;
  }

  T service<T extends ChopperService>(Type type) {
    final s = _services[type];
    if (s == null) {
      throw Exception("Service of type '$type' not found.");
    }
    return s as T;
  }

  void updateApiUrl(
      {String? apiUrl, String? apiProtocol, String? apiHost, String? apiPort}) {
    if (apiUrl?.isEmpty ?? false) {
      this.apiProtocol = apiProtocol ?? this.apiProtocol;
      this.apiHost = apiHost ?? this.apiHost;
      this.apiPort = apiPort ?? this.apiPort;
      this.apiUrl = apiProtocol! + apiHost! + apiPort!;
    } else {
      this.apiUrl = apiUrl!;
    }
  }

  ///Implement this method when checked what exceptions could be
  String? getLocalizedMessageFromException(dynamic exception) {
    if (exception == null) return null;
    if (exception is AdeptUtils.GenericException) {
      return getLocalizedMessageFromException(exception.causeException);
    }

    if (exception is WebSocketException ||
        exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException")) {
      return "Server connection timeout.";
    }
    if (exception is SocketException ||
        exception.toString().toLowerCase().contains("failed to connect")) {
      return "Check your internet connection. ${exception.toString()}";
    }
    if (exception is HandshakeException) return exception.toString();

    return exception.toString();
  }

  String? getGenericLocalizedMessageFromException(dynamic exception) {
    String? message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String? getLocalizedMessageFromHttpResultCode(int? code) {
    if (code == null) return null;
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";

      default:
        return "HTTP error code: ($code)";
    }
  }

  String? getLocalizedMessageFromResponse({Response? response, int? code}) {
    code ??= response?.statusCode;
    if (code == null) return null;
    String extraMessage = " ${response?.error ?? ""}";
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content.$extraMessage";

      case HttpStatus.badRequest:
        return "($code) Bad Request.$extraMessage";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized.$extraMessage";

      case HttpStatus.forbidden:
        return "($code) Forbidden.$extraMessage";

      case HttpStatus.notFound:
        return "($code) Not Found.$extraMessage";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error.$extraMessage";

      case AdeptUtils.Jsonable.jsonParserError:
        return "($code) Invalid response. JsonParser was unable to decode response.$extraMessage";

      default:
        return "HTTP error code: ($code).$extraMessage";
    }
  }
}
